import asyncio
from contextlib import asynccontextmanager
from typing import Optional

from ..forwarding import ResponseQueue, VirtualModelAdapter
from ..llms import ModelRunner
from ..llms.base import ModelResponseStream
from ..llms.types import ChatRequest, CompletionUsage
from ..logging import log
from ..threads import Session, Thread


class VirtualModelBase(ModelRunner):
    """Base class for virtual models.

    Virtual models usually add custom logic on top of the standard LLMs, and usually implement more complex flows,
    e.g. calling multiple models in sequence.

    The main method to implement is process(), which should write response chunks to the provided queue.
    """

    id: str

    async def run(self, model: str, request: ChatRequest, **kwargs) -> ModelResponseStream:
        queue = ResponseQueue()
        process_task = asyncio.create_task(self._process_wrapper(request, queue))
        return ModelResponseStream(VirtualModelAdapter(process_task, queue))

    async def _process_wrapper(self, request: ChatRequest, queue: ResponseQueue) -> None:
        try:
            await self.process(request, queue)
            log.debug(f"{self.__class__.__name__}._process_wrapper(): process finished")
        except Exception as error:
            log.critical(f"{self.__class__.__name__}.process() failed", error=error)
            await queue.add(f"Error: {repr(error)}")
        finally:
            await queue.put(None)

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        raise NotImplementedError()

    def get_assistant_messages(self, stream: ModelResponseStream) -> list[dict]:
        return [{"role": "assistant", "content": stream.text}]

    def cost_in_usd(self, model: str, usage: CompletionUsage) -> float:
        # TODO: not entirely correct, since we use sub-models, but can't track them ATM
        return 0

    async def process_continuation(
        self, session: Session, chat: ChatRequest, queue: ResponseQueue, model_name: str
    ) -> None:
        log.info("process_continuation()", model=self.id, messages_count=len(chat.messages))
        thread = session.create_thread(model_name, name="continuation")
        for message in chat.messages:
            thread.add_message(**message)

        async with thread.run() as stream:
            await stream.text_chunks().redirect(queue)

    @asynccontextmanager
    async def run_progressbar(self, queue: ResponseQueue, delay: float = 1.0, *, message: Optional[str] = None) -> None:
        """Runs a progress bar in the background, yielding dots to the queue.

        The progress bar is stopped when the context manager exits.
        """

        # If message is provided, output it first in italics
        if message:
            await queue.add(f"_{message}_ ")

        task = asyncio.create_task(self._run_progressbar_task(queue, delay))
        try:
            yield
        finally:
            task.cancel()
            await queue.add("\n")

    async def _run_progressbar_task(self, queue: ResponseQueue, delay: float) -> None:
        while True:
            await queue.add(".\u200b")
            await asyncio.sleep(delay)

    async def output_addendum(self, queue: ResponseQueue, session_or_thread: Session | Thread) -> None:
        await queue.add("\n\n---\n")
        await queue.add(f"Elapsed: {session_or_thread.elapsed_secs():.3f} secs\n")
        await queue.add("Costs: " + session_or_thread.cost_and_usage().pretty() + "\n")



import json
import random
from math import sqrt
from typing import Any

from ..tools_support.schema import jsonschema_for_function


def get_weather_at(lon: float, lat: float) -> dict[str, Any]:
    """
    Get the current weather in a given location. All units are in metric system.

    Args:
        lon (float): The longitude of the location.
        lat (float): The latitude of the location.

    Returns:
        Dict[str, Any]: The weather data. Contains the following fields:
            - temperature (float): The temperature in degrees Celsius.
            - humidity (float): The relative humidity.
            - wind_speed (float): The wind speed in meters per second.
            - weather (str): A description of the weather.
    """

    return {
        "temperature": random.uniform(-20, 40),
        "humidity": random.uniform(0, 100),
        "wind_speed": random.uniform(0, 20),
        "weather": random.choice(["sunny", "cloudy", "rainy", "snowy"]),
    }


async def get_location_coordinates(location: str) -> dict[str, float]:
    """
    Get the coordinates of a given location.

    Args:
        location (str): The name of the location.

    Returns:
        Dict[str, float]: The coordinates of the location. Contains the following fields:
            - lon (float): The longitude of the location.
            - lat (float): The latitude of the location.
    """

    return {
        "lon": random.uniform(-180, 180),
        "lat": random.uniform(-90, 90),
    }


def text_complexity(text: str) -> float:
    """Computes complexity score of the given text."""

    return sqrt(len(text))


if __name__ == "__main__":
    for func in [get_weather_at, get_location_coordinates]:
        schema = jsonschema_for_function(func)
        print(json.dumps(schema, indent=2))



import time
from typing import AsyncIterable

from fastapi import FastAPI, HTTPException, Request
from openai.types import CompletionUsage
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion import Choice as FullResponseChoice
from openai.types.chat.chat_completion_chunk import Choice, ChoiceDelta
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from sse_starlette import EventSourceResponse

from ..forwarding import ResponseQueue
from ..llms.types import ChatRequest
from ..logging import log
from ..threads import Session
from ..virtual_models.base import VirtualModelBase
from ..virtual_models.demos import DemoToolUsageModel, MinimalVirtualModel, SimpleVirtualModel
from ..virtual_models.multiphase import MultiphaseChatGPTModel, MultiphaseGPTModel, MultiphaseModel
from ..virtual_models.router import RouterGPTModel, RouterModel
from ..virtual_models.two_phase import TwoPhaseChatGPTModel, TwoPhaseGPTModel, TwoPhaseModel
from ..virtual_models.types import ModelCard, ModelList

app = FastAPI()


MODELS = [
    MinimalVirtualModel(),
    SimpleVirtualModel(),
    MultiphaseModel(),
    MultiphaseChatGPTModel(),
    MultiphaseGPTModel(),
    TwoPhaseModel(),
    TwoPhaseChatGPTModel(),
    TwoPhaseGPTModel(),
    RouterModel(),
    RouterGPTModel(),
    DemoToolUsageModel(),
]


class APIChatRequest(ChatRequest):
    model: str


@app.get("/v1/models")
def get_models() -> ModelList:
    return ModelList(data=[ModelCard(id=model.id, owned_by="Rivo") for model in MODELS])


async def process_wrapper(model: VirtualModelBase, chat: ChatRequest, queue: ResponseQueue) -> None:
    try:
        await model.process(chat, queue)
    finally:
        await queue.put(None)


async def chat_completion_streamed(model: VirtualModelBase, chat: ChatRequest) -> AsyncIterable[ChatCompletionChunk]:
    user_messages = [message for message in chat.messages if message["role"] == "user"]
    log.info(
        "app.chat_completion_streamed: starting",
        model=model.id,
        messages_count=len(chat.messages),
        user_messages_count=len(user_messages),
    )

    chat_id = "static"
    chunk_template = ChatCompletionChunk(
        id=chat_id,
        choices=[],
        created=int(time.time()),
        model=model.id,
        object="chat.completion.chunk",
    )

    yield chunk_template.model_copy(
        update={
            "choices": [Choice(delta=ChoiceDelta(content="", role="assistant"), index=0)],
        }
    )

    thread = Session().create_thread(model.id, chat, runner=model)
    async with thread.run_loop() as stream:
        async for chunk in stream.text_chunks():
            yield chunk_template.model_copy(
                update={
                    "choices": [Choice(delta=ChoiceDelta(content=chunk.content), index=0)],
                }
            )

    yield chunk_template.model_copy(
        update={
            "choices": [Choice(delta=ChoiceDelta(), finish_reason="stop", index=0)],
        }
    )
    yield chunk_template.model_copy(
        update={
            "choices": [],
            "usage": CompletionUsage(prompt_tokens=0, completion_tokens=0, total_tokens=0),
        }
    )

    log.info("app.chat_completion_streamed: done", model=model.id, cost=thread.cost_and_usage().pretty_root())


async def generate_chat_title(chat: ChatRequest, model_name: str = "gpt-4o") -> ChatCompletion:
    thread = Session().create_thread(model_name, ChatRequest(messages=chat.messages))
    async with thread.run() as stream:
        await stream.finish()

    title = stream.text
    cost_and_usage = thread.cost_and_usage()
    log.info("generate_chat_title()", title=title, usage=cost_and_usage.pretty_root())

    return ChatCompletion(
        id="static",
        choices=[
            FullResponseChoice(
                finish_reason="stop", index=0, message=ChatCompletionMessage(content=title, role="assistant")
            )
        ],
        created=int(time.time()),
        model=model_name,
        object="chat.completion",
        usage=CompletionUsage(
            prompt_tokens=cost_and_usage.prompt_tokens,
            completion_tokens=cost_and_usage.completion_tokens,
            total_tokens=cost_and_usage.total_tokens,
        ),
    )


@app.post("/v1/chat/completions", response_model=None)
async def post_chat_completions(request: Request, chat: APIChatRequest) -> EventSourceResponse | ChatCompletion:
    print(f"request: {request.method} {request.url.path}")

    model = next((m for m in MODELS if m.id == chat.model), None)
    if not model:
        raise HTTPException(status_code=404, detail="Unknown model")

    chat = ChatRequest(**chat.model_dump(exclude={"model"}))
    if not chat.stream:
        # Most likely WebUI request to generate chat title
        if "RESPOND ONLY WITH THE TITLE TEXT." in chat.messages[0]["content"]:
            return await generate_chat_title(chat)

        raise HTTPException(status_code=400, detail="Only streaming is supported")

    async def wrapper():
        async for chunk in chat_completion_streamed(model, chat):
            yield dict(data=chunk.model_dump_json())

    return EventSourceResponse(wrapper())



from typing import AsyncIterator, Callable, Optional

import anthropic
from anthropic import NOT_GIVEN, AsyncAnthropic, AsyncStream
from anthropic.types import ContentBlock, Message, RawMessageStreamEvent, TextBlock, ToolUseBlock, Usage

from ..logging import log
from ..tools_support.schema import jsonschema_for_function
from .base import APIConnectionError, ModelResponseStream, ModelRunner, RateLimitError, ResponseAdapterBase
from .types import (
    ChatRequest,
    CompletionUsage,
    ModelCost,
    ModelInfo,
    ResponseChunk,
    ResponseChunkToolCall,
    ResponseStopReasonChunk,
    ResponseTextChunk,
    ResponseUsageChunk,
    StopReason,
    ToolCall,
    ToolCallResult,
)


class AnthropicResponseAdapter(ResponseAdapterBase):
    def __init__(self, stream: AsyncStream[RawMessageStreamEvent]):
        super().__init__()

        self.stream = stream

        self.full_message: Optional[Message] = None
        self.raw_chunks = []

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        log.debug("AnthropicResponseAdapter.__aiter__()", id=id(self))
        if self.full_message is not None:
            log.error("AnthropicResponseAdapter.__aiter__(): already processed", id=id(self))
            return

        block: Optional[ContentBlock] = None
        tool_partial_json = ""

        async for chunk in self.stream:
            log.debug("AnthropicResponseAdapter.__aiter__(): got raw chunk", chunk=chunk)
            self.raw_chunks.append(chunk)

            if chunk.type == "message_start":
                assert self.full_message is None
                self.full_message = chunk.message
                yield ResponseUsageChunk(delta=self.parse_usage(chunk.message.usage))

            elif chunk.type == "message_stop":
                assert self.full_message is not None
                assert block is None

            elif chunk.type == "content_block_start":
                assert block is None
                block = chunk.content_block
                tool_partial_json = ""

            elif chunk.type == "content_block_stop":
                assert block is not None
                assert len(self.full_message.content) == chunk.index

                if isinstance(block, ToolUseBlock):
                    block.input = tool_partial_json
                    yield ResponseChunkToolCall(tool_call=ToolCall(id=block.id, name=block.name, arguments=block.input))

                self.full_message.content.append(block)
                block = None

            elif chunk.type == "content_block_delta":
                assert block is not None
                if chunk.delta.type == "text_delta":
                    assert isinstance(block, TextBlock)
                    assert chunk.delta.text
                    block.text += chunk.delta.text
                    yield ResponseTextChunk(content=chunk.delta.text)
                elif chunk.delta.type == "input_json_delta":
                    assert isinstance(block, ToolUseBlock)
                    tool_partial_json += chunk.delta.partial_json
                else:
                    raise ValueError(f"Unexpected block delta type: {chunk.delta.type}")

            elif chunk.type == "message_delta":
                if chunk.usage:
                    self.full_message.usage.output_tokens += chunk.usage.output_tokens
                    yield ResponseUsageChunk(delta=self.parse_usage(chunk.usage))
                if chunk.delta.stop_reason:
                    self.full_message.stop_reason = chunk.delta.stop_reason
                    reason = self.parse_stop_reason(chunk.delta.stop_reason)
                    yield ResponseStopReasonChunk(reason=reason)
                if chunk.delta.stop_sequence:
                    self.full_message.stop_sequence = chunk.delta.stop_sequence

            else:
                raise ValueError(f"Unexpected chunk type: {chunk.type}")

        log.debug("AnthropicResponseAdapter.__aiter__(): done", id=id(self))

    def parse_usage(self, usage: Usage) -> CompletionUsage:
        return CompletionUsage(prompt_tokens=getattr(usage, "input_tokens", 0), completion_tokens=usage.output_tokens)

    def parse_stop_reason(self, ant_reason: Optional[str]) -> Optional[StopReason]:
        if ant_reason == "end_turn":
            return StopReason.END_TURN
        if ant_reason == "tool_use":
            return StopReason.TOOL_USE
        if ant_reason == "max_tokens":
            return StopReason.LENGTH

        return None


class AnthropicRunner(ModelRunner):
    MODELS = [
        ModelInfo(
            name="claude-3-5-sonnet-20240620",
            aliases=["sonnet", "sonnet-3.5"],
            cost=ModelCost(prompt_per_1m=3, completion_per_1m=15),
        ),
    ]

    def __init__(self):
        super().__init__()
        self.client = AsyncAnthropic()

    async def run(self, model: str, request: ChatRequest, **kwargs) -> ModelResponseStream:
        self.log_run_request(model, request, kwargs)

        # Ensure the model is valid
        self.get_model_info(model)

        system_messages = [message for message in request.messages if message["role"] == "system"]
        assert len(system_messages) <= 1
        non_system_messages = [message for message in request.messages if message["role"] != "system"]

        try:
            stream = await self.client.messages.create(
                model=model,
                messages=non_system_messages,
                system=system_messages[0]["content"] if system_messages else NOT_GIVEN,
                tools=self.get_tools_schema(request.tools) if request.tools else NOT_GIVEN,
                tool_choice={"type": "auto"} if request.tools else NOT_GIVEN,
                max_tokens=4096,
                stream=True,
                **kwargs,
            )

        except anthropic.APIConnectionError as error:
            raise APIConnectionError(message=error.message) from error

        except anthropic.RateLimitError as error:
            metadata = {
                header: value
                for header, value in error.response.headers.items()
                if header.startswith("anthropic-ratelimit-")
            }
            # Flat 10-second retry for now
            raise RateLimitError(retry_in_secs=10, metadata=metadata) from error

        return ModelResponseStream(AnthropicResponseAdapter(stream))

    def get_tools_schema(self, tools: list[Callable]) -> list[dict]:
        return [jsonschema_for_function(func, parameters_key="input_schema") for func in tools]

    def get_tool_result_messages(self, results: list[ToolCallResult]) -> list[dict]:
        """Anthropic needs all tool results to be sent in a single message."""
        return [
            {
                "role": "user",
                "content": [
                    {
                        "type": "tool_result",
                        "tool_use_id": result.tool_call.id,
                        "is_error": result.is_error,
                        "content": result.content,
                    }
                    for result in results
                ],
            }
        ]

    def get_assistant_messages(self, stream: ModelResponseStream) -> list[dict]:
        if stream.stop_reason == StopReason.TOOL_USE:
            # Add assistant message with tool calls
            return [
                {
                    "role": "assistant",
                    "content": [
                        {
                            "type": "tool_use",
                            "id": tool_call.id,
                            "name": tool_call.name,
                            "input": tool_call.arguments_dict,
                        }
                        for tool_call in stream.get_tool_calls()
                    ],
                }
            ]
        else:
            # Add assistant message with completion
            return [{"role": "assistant", "content": stream.text}]



import inspect
import json
from typing import Any, Callable, Dict, Optional, get_args, get_origin

import typing_inspect
from docstring_parser import parse


def jsonschema_for_function(func: Callable, parameters_key: str = "parameters") -> Dict[str, Any]:
    """
    Generate a JSON schema for the given function.

    Args:
        func (Callable): The function to generate a schema for.
        parameters_key (str, optional): The key to use for the parameters. Defaults to "parameters".

    Returns:
        Dict[str, Any]: A dictionary representing the JSON schema.
    """
    schema = {
        "name": func.__name__,
        "description": "",
        parameters_key: {"type": "object", "properties": {}, "required": []},
    }

    # Parse the docstring
    docstring = parse(inspect.getdoc(func) or "")
    schema["description"] = docstring.short_description or ""

    # Get function signature
    sig = inspect.signature(func)

    for name, param in sig.parameters.items():
        param_schema = {}

        # Get type information
        if param.annotation != inspect.Parameter.empty:
            if typing_inspect.is_optional_type(param.annotation):
                type_name = _get_type_name(typing_inspect.get_args(param.annotation)[0])
            else:
                type_name = _get_type_name(param.annotation)

            if type_name.startswith("array["):
                item_type = type_name[6:-1]  # Extract inner type
                param_schema["type"] = "array"
                param_schema["items"] = {"type": item_type}
            else:
                param_schema["type"] = type_name

        # Get description from docstring
        param_doc = next((p for p in docstring.params if p.arg_name == name), None)
        if param_doc:
            param_schema["description"] = param_doc.description

        # Check if parameter is required
        if param.default == inspect.Parameter.empty:
            schema[parameters_key]["required"].append(name)

        schema[parameters_key]["properties"][name] = param_schema

    return schema


def _get_type_name(typ: Any) -> str:
    """Helper function to get a string representation of a type."""
    if get_origin(typ) == list:
        item_type = get_args(typ)[0]
        return f"array[{_get_type_name(item_type)}]"
    if typ == float:
        return "number"
    if typ == str:
        return "string"
    if typ == int:
        return "integer"
    if typ == bool:
        return "boolean"
    if typ == dict:
        return "object"
    if hasattr(typ, "__name__"):
        return typ.__name__.lower()
    return str(typ).lower()


if __name__ == "__main__":

    # Example usage:
    def get_current_weather(location: str, unit: Optional[str] = "celsius") -> Dict[str, Any]:
        """
        Get the current weather in a given location.

        Args:
            location (str): The city and state, e.g. San Francisco, CA
            unit (str, optional): The unit of temperature. Defaults to "celsius".

        Returns:
            Dict[str, Any]: A dictionary containing weather information.
        """
        pass

    schema = jsonschema_for_function(get_current_weather)
    print(json.dumps(schema, indent=2))



from typing import Optional

from prompt_toolkit import PromptSession
from prompt_toolkit.patch_stdout import patch_stdout

from aaf.event_handlers import use_event_handler
from aaf.threads import Thread
from aaf.ui import SimpleStreamingUserInterface


class ChatSession:
    """
    Manages an interactive chat session with an AI model.

    This class handles the user input loop, processes questions through the AI model,
    and manages the overall flow of the conversation, including handling user exit commands.

    Attributes:
        thread (Thread): The Thread object managing the conversation with the AI model.
        ui (SimpleStreamingUserInterface): The user interface for displaying output and handling input.
    """

    def __init__(self, thread: Thread, ui: Optional[SimpleStreamingUserInterface] = None):
        self.thread = thread
        self.ui = ui or SimpleStreamingUserInterface()
        self.prompt_session = PromptSession()

    async def run_loop(self, initial_question: Optional[str] = None, *, interactive: bool = True):
        if initial_question:
            await self._process_question(initial_question)

        try:
            while interactive:
                question = await self._get_user_input()
                if question is None:
                    break
                await self._process_question(question)

        finally:
            self.ui.info(f"Final cost: {self.thread.cost_and_usage().pretty_root()}")

    async def _get_user_input(self) -> Optional[str]:
        with patch_stdout():
            try:
                question = await self.prompt_session.prompt_async("You: ")
                if question.lower() in ("exit", "quit"):
                    return None
                return question
            except EOFError:
                # Handle Ctrl+D
                print("\nGoodbye!")
                return None

    async def _process_question(self, question: str):
        self.thread.add_message("user", question)
        async with self.thread.run_loop() as stream:
            await use_event_handler(stream, self.ui)

        self.ui.info(f"Current cost: {self.thread.cost_and_usage().pretty_root()}")



from typing import Optional

from .. import prompts
from ..forwarding import ResponseQueue
from ..llms.types import ChatRequest
from ..logging import log
from ..prompts import MULTIPHASE_FEEDBACK, MULTIPHASE_FINAL
from ..threads import Session
from ..utils import extract_xml_fragment
from .base import VirtualModelBase


class MultiphaseModel(VirtualModelBase):
    id: str = "rivo/multiphase"

    external_model_name = "sonnet-3.5"

    def __init__(self, external_model_name: Optional[str] = None):
        super().__init__()

        if external_model_name is not None:
            self.external_model_name = external_model_name

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        session = Session()

        user_messages = [message for message in chat.messages if message["role"] == "user"]
        if len(user_messages) > 1:
            return await self.process_continuation(session, chat, queue, self.external_model_name)

        question = user_messages[0]["content"]

        prompt = await self.generate_prompt(session, queue, question)
        await queue.add_debug(f"## _Multiphase: prompt_\n\n{prompt}\n\n---\n\n")

        draft = await self.phase_one(session, queue, prompt, question)
        await queue.add_debug(f"## _Multiphase: draft answer_\n\n{draft}\n\n---\n\n")

        feedback = await self.phase_two(session, queue, question, draft)
        await queue.add_debug(f"## _Multiphase: feedback_\n\n{feedback}\n\n---\n\n")

        answer = await self.phase_three(session, queue, prompt, question, draft, feedback)

        await self.output_addendum(queue, session)

    async def generate_prompt(self, session: Session, queue: ResponseQueue, question: str) -> str:
        thread = session.create_thread(self.external_model_name, name="prompt")
        thread.add_message("system", prompts.PROMPT_ENGINEER)
        thread.add_message("user", question)

        async with self.run_progressbar(queue, message="Writing prompt"):
            async with thread.run() as stream:
                await stream.finish()

        full_response = thread.messages[-1]["content"]
        log.debug("Multiphase.generate_prompt(): Full response", response=full_response)

        result = extract_xml_fragment(full_response, "system_prompt")
        log.info(
            "MultiphaseModel.generate_prompt() done", result_len=len(result), cost=thread.cost_and_usage().as_log()
        )
        return result

    async def phase_one(self, session: Session, queue: ResponseQueue, prompt: str, question: str) -> str:
        """Asks the user's question and returns initial answer."""

        temperature = 0.7

        thread = session.create_thread(self.external_model_name, name="draft")
        thread.add_message("system", prompt)
        thread.add_message("user", question)

        async with self.run_progressbar(queue, message="Drafting initial answer"):
            async with thread.run(temperature=temperature) as stream:
                await stream.finish()

        result = thread.messages[-1]["content"]
        log.info("MultiphaseModel.phase_one() done", result_len=len(result), cost=thread.cost_and_usage().as_log())
        return result

    async def phase_two(self, session: Session, queue: ResponseQueue, question: str, initial_answer: str) -> str:
        """Writes feedback for initial answer."""

        temperature = 1.0

        thread = session.create_thread(self.external_model_name, name="feedback")
        thread.add_message("system", MULTIPHASE_FEEDBACK)
        thread.add_message(
            "user",
            f"""
            <question>\n{question}\n</question>\n
            <answer>\n{initial_answer}\n</answer>\n
        """,
        )
        async with self.run_progressbar(queue, message="Looking for ways to improve"):
            async with thread.run(temperature=temperature) as stream:
                await stream.finish()

        result = thread.messages[-1]["content"]
        log.info("MultiphaseModel.phase_two() done", result_len=len(result), cost=thread.cost_and_usage().as_log())
        return result

    async def phase_three(
        self, session: Session, queue: ResponseQueue, prompt: str, question: str, initial_answer: str, feedback: str
    ) -> str:
        """Composes final answer based on initial answer and feedback."""

        temperature = 0.7

        thread = session.create_thread(self.external_model_name, name="answer")
        thread.add_message("system", prompt + MULTIPHASE_FINAL)
        thread.add_message(
            "user",
            f"""
            <question>\n{question}\n</question>\n
            <answer>\n{initial_answer}\n</answer>\n
            <feedback>\n{feedback}\n</feedback>\n
            """,
        )

        async with thread.run(temperature=temperature) as stream:
            # Display progress bar until the first "-----" is encountered, marking the start of the final answer
            end_of_thinking_seen = False
            async with self.run_progressbar(queue, message="Composing final answer"):
                text = ""
                async for chunk in stream.text_chunks():
                    text += chunk.content

                    if "</thinking>" in text:
                        end_of_thinking_seen = True
                        break

            # Display the final answer
            await queue.add("\n")
            async for chunk in stream.text_chunks():
                await queue.put(chunk)

            # Failsafe, in case the model forgot the </thinking> tag. Just show the full response
            if not end_of_thinking_seen:
                log.info("Multiphase: End of thinking not seen, showing full response")
                await queue.add(text)

        result = thread.messages[-1]["content"]
        log.info("MultiphaseModel.phase_three() done", result_len=len(result), cost=thread.cost_and_usage().as_log())
        return result


class MultiphaseGPTModel(MultiphaseModel):
    id: str = "rivo/multiphase-gpt"

    external_model_name = "gpt-4o"


class MultiphaseChatGPTModel(MultiphaseModel):
    id: str = "rivo/multiphase-chatgpt"

    external_model_name = "chatgpt-4o-latest"



import json
from asyncio import Queue
from enum import Enum
from typing import Any, AsyncIterator, Callable, Optional

from pydantic import BaseModel

from ..logging import log


class CompletionUsage(BaseModel):
    completion_tokens: int = 0
    """Number of tokens in the generated completion."""

    prompt_tokens: int = 0
    """Number of tokens in the prompt."""

    @property
    def total_tokens(self):
        """Total number of tokens used in the request (prompt + completion)."""
        return self.prompt_tokens + self.completion_tokens

    def __add__(self, other: "CompletionUsage") -> "CompletionUsage":
        return CompletionUsage(
            completion_tokens=self.completion_tokens + other.completion_tokens,
            prompt_tokens=self.prompt_tokens + other.prompt_tokens,
        )


class CostAndUsage(CompletionUsage):
    cost: float = 0
    name: str = "_unknown_"

    children: list["CostAndUsage"] = []

    @classmethod
    def aggregate(cls, name: str, children: list["CostAndUsage"]) -> "CostAndUsage":
        total = cls(name=name, children=children)
        for child in children:
            total.completion_tokens += child.completion_tokens
            total.prompt_tokens += child.prompt_tokens
            total.cost += child.cost

        return total

    def pretty_root(self) -> str:
        return f"{self.name}: {self.cost:.4f} USD  ({self.prompt_tokens} + {self.completion_tokens} tokens)"

    def pretty(self, *, level: int = 0) -> str:
        result = self.pretty_root()
        if level:
            result = f"{(level - 1) * '    '}- {result}"

        for child in self.children:
            result += "\n" + child.pretty(level=level + 1)
        return result

    def as_log(self) -> str:
        return f"{self.cost:.4f} USD; {self.prompt_tokens} + {self.completion_tokens} tokens"


class ToolCall(BaseModel):
    id: str

    name: str
    """The name of the function to call."""

    arguments: str
    """
    The arguments to call the function with, as generated by the model in JSON
    format. Note that the model does not always generate valid JSON, and may
    hallucinate parameters not defined by your function schema. Validate the
    arguments in your code before calling your function.
    """

    @property
    def arguments_dict(self) -> dict[str, Any]:
        return json.loads(self.arguments)

    def pretty(self) -> str:
        return f"{self.name}({', '.join(f'{k}={repr(v)}' for k, v in self.arguments_dict.items())})"


class ToolCallResult(BaseModel):
    tool_call: ToolCall
    is_error: bool = False
    content: Optional[str] = None


# TODO: use StrEnum when we upgrade to Python 3.11
class StopReason(Enum):
    END_TURN = "end_turn"
    """The conversation has ended."""

    TOOL_USE = "tool_use"
    """A tool was used."""

    LENGTH = "length"
    """The maximum length was reached."""


class ResponseChunk(BaseModel):
    pass


class ResponseBaseTextChunk(ResponseChunk):
    content: str


class ResponseTextChunk(ResponseBaseTextChunk):
    pass


class ResponseVerboseChunk(ResponseBaseTextChunk):
    pass


class ResponseDebugChunk(ResponseBaseTextChunk):
    pass


class ResponseChunkToolCall(ResponseChunk):
    tool_call: ToolCall


class ResponseChunkToolCallStarted(ResponseChunk):
    tool_call: ToolCall


class ResponseChunkToolCallFinished(ResponseChunk):
    tool_call: ToolCall
    result: Any


class ResponseChunkToolCallFailed(ResponseChunk):
    tool_call: ToolCall
    error: str


class ResponseUsageChunk(ResponseChunk):
    delta: CompletionUsage


class ResponseStopReasonChunk(ResponseChunk):
    reason: StopReason


class ResponseChunkCompleteText(ResponseChunk):
    content: str


class ResponseControlChunk(BaseModel):
    pass


class ResponseControlChunkStreamBegin(BaseModel):
    pass


class ResponseControlChunkStreamEnd(BaseModel):
    pass


class ResponseControlChunkRateLimited(BaseModel):
    delay_secs: float
    metadata: dict


class ResponseChunksStream:

    def __init__(self, stream: "ResponseStream", filter: Callable[[ResponseChunk], bool]):
        self.stream = stream
        self.filter = filter

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        async for chunk in self.stream:
            if self.filter(chunk):
                yield chunk

    async def redirect(self, queue: Queue) -> None:
        async for chunk in self:
            await queue.put(chunk)


class ResponseStream:
    """Represents a stream of response chunks from model.

    Stream contains either "physical" response (single model completion, from an API call) or a "logical" response
    (chain of responses, e.g. with tool calls).

    It also has a few helper methods.
    """

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        raise NotImplementedError()

    def all_chunks(self) -> ResponseChunksStream:
        return ResponseChunksStream(self, lambda chunk: True)

    def text_chunks(self) -> ResponseChunksStream:
        return ResponseChunksStream(self, lambda chunk: isinstance(chunk, ResponseTextChunk))

    async def redirect(self, queue: Queue) -> None:
        await self.all_chunks().redirect(queue)

    async def finish(self) -> None:
        async for _ in self:
            pass

    async def put(self, chunk: ResponseChunk) -> None:
        raise NotImplementedError()

    @property
    def text(self) -> str:
        raise NotImplementedError()

    @property
    def usage(self) -> CompletionUsage:
        raise NotImplementedError()


class ModelCost(BaseModel):
    prompt_per_1m: float = 0
    completion_per_1m: float = 0


class ModelInfo(BaseModel):
    name: str
    aliases: list[str] = []
    cost: ModelCost = ModelCost()


class ChatRequest(BaseModel):
    messages: list[dict] = []
    stream: bool = True
    max_tokens: Optional[int] = None
    tools: list[Callable] = []
    tool_choice: str | dict = "auto"



import time
from typing import Literal

from pydantic import BaseModel, Field


class ModelCard(BaseModel):
    id: str
    object: Literal["model"] = "model"
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: str = ""


class ModelList(BaseModel):
    object: Literal["list"] = "list"
    data: list[ModelCard] = []



PROMPT_ENGINEER = """
Act as an expert prompt engineer for large language models.
Use the best prompt engineering practices applications to latest models such as GPT-4 and Anthropic Claude 3 families.
The prompt should be usable as system prompt, but should also fit as beginning of the initial user message.

Some of the overall prompting recommendations (from OpenAI and others) are:
- Ask the model to adopt a specific persona who is a subject matter expert.
- Delimiters like triple quotation marks, XML tags, section titles, etc. can help demarcate sections of text to be treated differently.
- Some tasks are best specified as a sequence of steps. Writing the steps out explicitly can make it easier for the model to follow them.
- Providing general instructions that apply to all examples is generally more efficient than demonstrating all permutations of a task by example, but in some cases providing examples may be easier. For example, if you intend for the model to copy a particular style of responding to user queries which is difficult to describe explicitly. This is known as "few-shot" prompting.
- You can ask the model to produce outputs that are of a given target length. The targeted output length can be specified in terms of the count of words, sentences, paragraphs, bullet points, etc. 
- Let the model know that they can ask follow up questions if something is unclear. In that case, they should not proceed with the final answer before user has clarified the uncertain parts.
- Don't force the model to always answer, e.g: If the answer cannot be found in the articles, write "I could not find an answer."
- Strategy: Give models time to "think"
    - Instruct the model to work out its own solution before rushing to a conclusion
    - Use inner monologue or a sequence of queries to hide the model's reasoning process

You can include those fragments, verbatim, proven to make the LLMs perform better:
- Explain it with gradually increasing complexity.
- Before you answer, make a list of wrong assumptions people sometimes make about the concepts included in the question.
- Let's work this out in a step by step way to be sure we have the right answer.
- Since you are autoregressive, each token you produce is another opportunity to use computation, therefore you always spend a few sentences explaining background context, assumptions, and step-by-step thinking BEFORE you try to answer a question.

The prompt should explicitly encourage the model to:
- First provide some internal monologue and reasoning, followed by the full answer.
- Answer in Markdown format, e.g with headers and bold/italic, unless user explicitly requested otherwise.
- They can make changes to the suggested answer structure, when it makes sense.

Include the contents of following block (without backticks) in the prompt verbatim, with no changes:
```
Begin your response with your internal thought process, enclosed in <thinking>...</thinking> tags, followed by the full, structured answer to the user's question.
```

Keeping the above in mind, generate prompt for the user's request. Do NOT attempt to answer the request, only generate
the prompt that will be use to generate the best possible answer.
The prompt does not need to contain the request itself, as it will be given to the model in a separate message.

Before generating the prompt, follow the advice above and give yourself time to think,
by writing out some thoughts and reasoning.
After than, write out the prompt.

The format of your response should be as follows:

<thinking>
Here should be your internal monologue with thoughts and reasoning.
</thinking>
<system_prompt>
Here should be the final prompt for the LLM (Large Language Model) that will answer the user's request.
</system_prompt>
""".strip()


MULTIPHASE_DRAFT = """
Carefully analyze the question from user and provide a detailed analysis and response.
In the answer:
1. include your internal monologue, taking time to consider the question.
2. analyze the various options, when applicable, and describe them in detail.
3. provide conclusion, with a clear recommendation or summary of the analysis.

Answer in Markdown format, e.g with headers and bold/italic, unless user explicitly requested otherwise.
"""

# Generated by:
# python prompt_engineer.py \
# "You are part two of three in a multiphase question answering chain.
# The first part composes a draft answer, second part analyses it and gives feedback for improvements,
# and third part composes final answer based on question, draft, and feedback.
# Your task is to analyze the question and the answer provided, and give constructive feedback,
# focusing on potential improvements."

MULTIPHASE_FEEDBACK = """
You are an expert editor and reviewer, tasked with analyzing draft answers and providing constructive feedback for improvements.
You are the second part of a three-phase question-answering system. Your role is crucial in ensuring the final answer is of the highest quality.

For each task, you will receive the original question and a draft answer. Your job is to:

1. Carefully analyze both the question and the draft answer.
2. Identify strengths in the draft answer, highlighting what has been done well.
3. Pinpoint areas for improvement, considering factors such as:
   - Accuracy of information
   - Completeness of the answer
   - Clarity and coherence
   - Relevance to the original question
   - Depth of explanation
   - Use of examples or evidence to support claims
   - Logical flow and structure

4. Provide specific, actionable feedback for each area of improvement. Explain why these changes are necessary and how they will enhance the answer.

5. If applicable, suggest additional points or perspectives that could be included to make the answer more comprehensive.

6. Consider the target audience and ensure the feedback aligns with their needs and level of understanding.

7. If you notice any potential misconceptions or common errors related to the topic, point them out to be addressed in the final answer.

Format your response as follows:

```markdown
## Strengths
- [List the strong points of the draft answer]

## Areas for Improvement
1. [First area of improvement]
   - Explanation of the issue
   - Suggestion for improvement
   - Example or clarification if needed

2. [Second area of improvement]
   - Explanation of the issue
   - Suggestion for improvement
   - Example or clarification if needed

[Continue with additional points as needed]

## Additional Suggestions
- [Any extra points or perspectives to consider]

## Overall Recommendation
[Summarize your main points and give a general recommendation for improving the answer]
```

Remember, your feedback will be used by the third part of the system to create the final answer. Be thorough, constructive, and clear in your analysis to ensure the best possible outcome.

If any part of the question or draft answer is unclear, state this in your response and explain what clarification is needed before proceeding with a full analysis.

At the end of feedback, provide a score for the draft answer in the range 0-10. 0 should indicate a very bad answer,
5 is quite fine but could use improvement, 8 is one that could already directly be sent to user, and 10 is ideal.
Provide it in xml tag: <score>6</score>
"""

MULTIPHASE_FINAL = """
Update: draft of the answer has already been written, and we have feedback for it.
Your task is to take into account the feedback, and provide a revised answer to the user's question.

Provide some internal monologue and analysis in <thinking>...</thinking> tags, followed by the full revised answer. 
DO NOT skip any sections of the revised answer, even if they remain unchanged. Do not include any internal
comments outside of <thinking> tags - the final answer must be exactly what will be provided to the user.
"""



from ..forwarding import ResponseQueue
from ..llms.types import ChatRequest, ResponseControlChunkRateLimited
from ..logging import log
from ..threads import Session
from ..utils import extract_xml_fragment
from .base import VirtualModelBase

ROUTER_PROMPT = """
You will receive user's request and must categorize it, selecting which model to send it to.
Do NOT attempt to fulfill the request or answer it.

Here are the models available:
- simple: this should be used for brief or simple requests.
- pirate: use this if user mentions wanting the response to be in pirate speak, or if the request is strongly related to pirates.
- two-phase: use this for more complex requests, that need deeper thinking and advanced models.
- multiphase: this is for the most complex, research-like tasks, where the answer should be long and comprehensive.

Begin your response with your internal thought process, enclosed in <thinking>...</thinking> tags.
This must be followed by <model>[model name]</model> where [model name] is one of the models listed above.

Here are some examples:
<example>
<question>What is the capital of France?</question>
<response>
<thinking>This is a simple request, the answer is factual and user did not request elaboration.</thinking>
<model>simple</model>
</response>
</example>

<example>
<question>What are the districts of the capital of France, and which one would you recommend for a visit?</question>
<response>
<thinking>This is a more complex request, requiring detailed information and analysis.</thinking>
<model>two-phase</model>
</response>
</example>

<example>
<question>What are the main causes of climate change and how can they be mitigated?</question>
<response>
<thinking>This is a complex, research-like question that requires a comprehensive answer.</thinking>
<model>multiphase</model>
</response>
</example>
"""


class RouterModel(VirtualModelBase):
    id: str = "rivo/router"

    external_model_name = "sonnet-3.5"

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        session = Session()

        user_messages = [message for message in chat.messages if message["role"] == "user"]
        if len(user_messages) > 1:
            return await self.process_continuation(session, chat, queue, self.external_model_name)

        question = user_messages[0]["content"]

        model_type = await self.select_model(session, queue, question)
        await queue.add(f"_Router: model: {model_type}_\n\n")

        # TODO: call the selected model

        await self.output_addendum(queue, session)

    async def select_model(self, session: Session, queue: ResponseQueue, question: str) -> str:
        thread = session.create_thread(self.external_model_name, name="model selection")
        thread.add_message("system", ROUTER_PROMPT)
        thread.add_message("user", question)

        async with self.run_progressbar(queue, message="Choosing model"):
            async with thread.run() as stream:
                async for chunk in stream:
                    if isinstance(chunk, ResponseControlChunkRateLimited):
                        await queue.add(" _(rate limited)_ ")
                await stream.finish()

        full_response = thread.messages[-1]["content"]
        log.info(f"Router: model selection response:\n{full_response}")

        return extract_xml_fragment(full_response, "model")


class RouterGPTModel(RouterModel):
    id: str = "rivo/router-gpt"

    external_model_name = "gpt-4o"



import asyncio
from typing import AsyncIterator, Optional

from .. import logging
from ..logging import log
from ..utils import truncate_text
from .types import (
    ChatRequest,
    CompletionUsage,
    ModelInfo,
    ResponseChunk,
    ResponseChunkCompleteText,
    ResponseChunkToolCall,
    ResponseControlChunkStreamBegin,
    ResponseControlChunkStreamEnd,
    ResponseStopReasonChunk,
    ResponseStream,
    ResponseTextChunk,
    ResponseUsageChunk,
    StopReason,
    ToolCall,
    ToolCallResult,
)


class RateLimitError(Exception):
    def __init__(self, retry_in_secs: float, metadata: dict):
        super().__init__()

        self.retry_in_secs = retry_in_secs
        self.metadata = metadata


class APIConnectionError(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return f"{self.__class__.__name__}(message={self.message!r})"


class ResponseAdapterBase:
    """Base class for adapters which translate model-specific response stream into a common format.

    It should take model-specific response and yield following ResponseChunk instances:
    - ResponseTextChunk (additive)
    - ResponseUsageChunk (additive)
    - ResponseChunkToolCall
    - ResponseStopReasonChunk

    Chunks should be emitted as soon as they are available, in the original order.
    All Text and ToolCall chunks should be emitted before the StopReason chunk.
    """

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        raise NotImplementedError()


class ModelResponseStream(ResponseStream):
    """Stateful response stream from a model.

    It works with adapter to get basic response stream from model, and adds some higher-level stateful logic.

    For example, it keeps track of full response text, tool calls, stop reason, and token usage.
    It also adds some control-flow chunks to the response stream, e.g for detecting beginning or end of the stream.

    It supports pausing and then continuing processing of the stream, e.g across multiple for loops.
    """

    def __init__(self, adapter: ResponseAdapterBase):
        super().__init__()

        self._text = ""
        self._usage = CompletionUsage()
        self._stop_reason = None
        self._is_finished = False
        self._tool_calls = []
        self._exception = None

        self.processed_queue = asyncio.Queue()
        self.processed_chunks: list[ResponseChunk] = []

        self.run_adapter_task = asyncio.create_task(self.background_process(adapter, self.processed_queue))

    @property
    def text(self) -> str:
        return self._text

    @property
    def usage(self) -> CompletionUsage:
        return self._usage

    @property
    def stop_reason(self) -> Optional[StopReason]:
        return self._stop_reason

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        if self._is_finished:
            return

        while True:
            chunk = await self.processed_queue.get()
            if chunk is None:
                self._is_finished = True
                break

            self.processed_chunks.append(chunk)

            if isinstance(chunk, Exception):
                self._exception = chunk

            yield chunk

    async def background_process(self, adapter: ResponseAdapterBase, queue: asyncio.Queue) -> None:
        """The main method for iterating over the response stream.

        It's implemented as wrapper around iterate_raw_stream(). The latter is responsible for emitting 'raw' chunks
        from the stream.
        This method adds some generic, higher-level logic for flow control and state management.
        """

        if self._is_finished:
            return

        log.debug("ModelResponseStream.background_process()", stream_id=id(self))
        await queue.put(ResponseControlChunkStreamBegin())

        async for chunk in adapter:
            log.debug("ModelResponseStream.background_process(): got chunk", stream_id=id(self), chunk=chunk)
            if isinstance(chunk, ResponseTextChunk):
                self._text += chunk.content
            if isinstance(chunk, ResponseUsageChunk):
                self._usage += chunk.delta
            if isinstance(chunk, ResponseChunkToolCall):
                self._tool_calls.append(chunk.tool_call)

            if isinstance(chunk, ResponseStopReasonChunk):
                # End of stream
                self._stop_reason = chunk.reason

                if self._text:
                    await queue.put(ResponseChunkCompleteText(content=self._text))

            await queue.put(chunk)
            log.debug("ModelResponseStream.background_process(): added to queue", stream_id=id(self), chunk=chunk)

        await queue.put(ResponseControlChunkStreamEnd())
        await queue.put(None)
        log.debug("ModelResponseStream.background_process(): DONE", stream_id=id(self))

    async def finish(self, ignore_exceptions: bool = False) -> None:
        log.debug("ModelResponseStream.finish(): finishing...", stream_id=id(self))
        await self.run_adapter_task

        # Consume the stream
        async for _ in self:
            pass

        if self._exception is not None and not ignore_exceptions:
            log.error("ModelResponseStream.finish(): exception", stream_id=id(self), exception=self._exception)
            raise self._exception

        log.debug("ModelResponseStream.finish(): done", stream_id=id(self))

    def get_tool_calls(self) -> list[ToolCall]:
        return self._tool_calls


class ModelRunner:
    """Invokes model with given messages and parameters."""

    # Lists all supported models. Each element is (model_name, aliases).
    MODELS: list[ModelInfo]

    def __init__(self):
        pass

    async def run(self, model: str, request: ChatRequest, **kwargs) -> ModelResponseStream:
        raise NotImplementedError()

    def cost_in_usd(self, model: str, usage: CompletionUsage) -> float:
        cost_info = self.get_model_info(model).cost
        return (
            usage.prompt_tokens * cost_info.prompt_per_1m / 1_000_000
            + usage.completion_tokens * cost_info.completion_per_1m / 1_000_000
        )

    @classmethod
    def get_provider_and_model(
        cls, providers: list[type["ModelRunner"]], model_name: str
    ) -> tuple[type["ModelRunner"], str]:
        """Resolves the user-specified model name into a provider and a canonical model name."""

        for provider in providers:
            for info in provider.MODELS:
                if model_name in [info.name] + info.aliases:
                    return provider, info.name

        raise ValueError(f"Model {model_name} is not supported by any of the providers: {providers}")

    @classmethod
    def get_model_info(cls, model: str) -> ModelInfo:
        for model_info in cls.MODELS:
            if model in [model_info.name] + model_info.aliases:
                return model_info
        else:
            raise ValueError(f"Model {model} is not supported by {cls.__name__}")

    def create_tool_result_message(self, tool_call: ToolCall, result: str) -> dict:
        raise NotImplementedError()

    def create_tool_error_message(self, tool_call: ToolCall) -> dict:
        raise NotImplementedError()

    def get_tool_result_messages(self, results: list[ToolCallResult]) -> list[dict]:
        raise NotImplementedError()

    def get_assistant_messages(self, stream: ModelResponseStream) -> list[dict]:
        raise NotImplementedError()

    def log_run_request(self, model: str, request: ChatRequest, kwargs: dict) -> None:
        event = f"{self.__class__.__name__}.run()"
        if logging.min_level <= logging.DEBUG:
            log.debug(event, model=model, request=request, kwargs=kwargs)
        else:
            # Create a version of messages with content truncated, to make the log message more manageable
            brief_request = request.model_copy(
                update={"messages": self.truncate_messages_for_logging(request.messages)}
            )
            log.info(event, model=model, request=brief_request, kwargs=kwargs)

    def truncate_messages_for_logging(self, messages: list[dict]) -> list[dict]:
        truncated = []
        for full_message in messages:
            message = full_message.copy()
            if isinstance(message["content"], str):
                message["content"] = truncate_text(message["content"], 25)
            if message.get("tool_calls"):
                message["tool_calls"] = f"<{len(message['tool_calls'])} tool call requests>"
            truncated.append(message)

        return truncated



import logging
from datetime import datetime

import rich
from rich.markup import escape

CRITICAL = logging.CRITICAL
FATAL = logging.FATAL
ERROR = logging.ERROR
WARNING = logging.WARNING
WARN = logging.WARN
INFO = logging.INFO
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET

min_level: int = INFO


def set_min_level(level: int) -> None:
    global min_level
    min_level = level
    logging.basicConfig(level=level)


def set_level_from_flags(*, quiet: bool = False, debug: bool = False):
    if debug:
        set_min_level(DEBUG)
    elif quiet:
        set_min_level(WARNING)


class Logger:
    def log(self, level: int, event: str, **kwargs) -> None:
        if level < min_level:
            return

        timestamp = datetime.now().isoformat(" ", timespec="milliseconds")
        level_name = logging.getLevelName(level)
        attributes = " ".join(f"{k}={repr(v)}" for k, v in kwargs.items())
        console = rich.get_console()
        console.print(escape(f"{timestamp}: [{level_name}] {event}  {attributes}"), style="dim", highlight=False)

        if level >= CRITICAL:
            console.print_exception(show_locals=True)

    def debug(self, event: str, **kwargs) -> None:
        self.log(logging.DEBUG, event, **kwargs)

    def info(self, event: str, **kwargs) -> None:
        self.log(logging.INFO, event, **kwargs)

    def warning(self, event: str, **kwargs) -> None:
        self.log(logging.WARNING, event, **kwargs)

    def error(self, event: str, **kwargs) -> None:
        self.log(logging.ERROR, event, **kwargs)

    def critical(self, event: str, **kwargs) -> None:
        self.log(logging.CRITICAL, event, **kwargs)


log = Logger()



from .anthropic import AnthropicRunner
from .base import ModelRunner
from .ollama import OllamaRunner
from .openai import OpenAIRunner

PROVIDERS = [OpenAIRunner, AnthropicRunner, OllamaRunner]


def get_llm_provider_and_model(model_name: str) -> tuple[type["ModelRunner"], str]:
    return ModelRunner.get_provider_and_model(PROVIDERS, model_name)



def extract_xml_fragment(source: str, tag: str) -> str:
    start_tag = f"<{tag}>"
    end_tag = f"</{tag}>"

    try:
        start = source.index(start_tag) + len(start_tag)
    except ValueError:
        return source

    try:
        end = source.index(end_tag, start)
    except ValueError:
        end = len(source)

    return source[start:end]


def truncate_text(text: str, max_length: int) -> str:
    if len(text) > max_length:
        return text[: max_length - 3] + "..."
    return text



import rich
from rich.text import Text

from .event_handlers import EventHandlerBase
from .llms.types import ResponseStream, ResponseTextChunk, ToolCall
from .utils import truncate_text


class SimpleStreamingUserInterface(EventHandlerBase):
    """Simple UI that shows streamed responses and some status info."""

    stream_responses = True

    def info(self, message: str):
        rich.print(f"[bold]{message}[/bold]")

    def debug(self, message: str):
        print(f"{message}")

    def assistant_message_stream_start(self, stream: ResponseStream):
        pass

    def assistant_message_stream_chunk(self, stream: ResponseStream, chunk: ResponseTextChunk):
        if self.stream_responses:
            text = Text(chunk.content, style="bold yellow")
            rich.print(text, end="")

    def assistant_message_stream_end(self, stream: ResponseStream):
        if self.stream_responses:
            rich.print()
        else:
            text = Text(stream.text, style="bold yellow")
            rich.print(text)

    def tool_call_start(self, tool_call: ToolCall):
        rich.print(f"Starting tool call: [bold]{tool_call.pretty()}[/bold]")

    def tool_call_result(self, tool_call: ToolCall, result: str):
        print(f"    result: {truncate_text(str(result), 200)}")

    def tool_call_error(self, tool_call: ToolCall, error: str):
        rich.print(f"    [bold red]error: {error}[/bold red]")



from .llms.types import (
    ResponseChunkToolCallFailed,
    ResponseChunkToolCallFinished,
    ResponseChunkToolCallStarted,
    ResponseControlChunkStreamBegin,
    ResponseControlChunkStreamEnd,
    ResponseStream,
    ResponseTextChunk,
    ToolCall,
)


class EventHandlerBase:
    """Base class for event handlers that process messages and events from the assistant and tools.

    This can be used to implement User Interface classes, custom logic, etc.
    """

    def info(self, message: str):
        """Print an informational message."""
        pass

    def debug(self, message: str):
        """Print a debug message."""
        pass

    def assistant_message_stream_start(self, stream: ResponseStream):
        """Called when the assistant starts sending a message stream."""
        pass

    def assistant_message_stream_chunk(self, stream: ResponseStream, chunk: ResponseTextChunk):
        """Called for each chunk of the assistant's response."""
        pass

    def assistant_message_stream_end(self, stream: ResponseStream):
        """Called when the assistant finishes sending a message stream."""
        pass

    def tool_call_start(self, tool_call: ToolCall):
        """Called when a tool call is about to be executed."""
        pass

    def tool_call_result(self, tool_call: ToolCall, result: str):
        """Called when a tool call has been executed successfully."""
        pass

    def tool_call_error(self, tool_call: ToolCall, error: str):
        """Called when a tool call has failed."""
        pass

    def loop_end(self, cost: float):
        pass


async def use_event_handler(stream: ResponseStream, handler: EventHandlerBase):
    """Uses the event handler for the stream's messages.

    This is basically an adapter that directs the stream's messages to the event handler.
    """

    async for chunk in stream.all_chunks():
        if isinstance(chunk, ResponseControlChunkStreamBegin):
            handler.assistant_message_stream_start(stream)
        elif isinstance(chunk, ResponseControlChunkStreamEnd):
            handler.assistant_message_stream_end(stream)
        elif isinstance(chunk, ResponseTextChunk):
            handler.assistant_message_stream_chunk(stream, chunk)
        elif isinstance(chunk, ResponseChunkToolCallStarted):
            handler.tool_call_start(chunk.tool_call)
        elif isinstance(chunk, ResponseChunkToolCallFinished):
            handler.tool_call_result(chunk.tool_call, chunk.result)
        elif isinstance(chunk, ResponseChunkToolCallFailed):
            handler.tool_call_error(chunk.tool_call, chunk.error)



import asyncio
from typing import Annotated, Callable

import rich
import typer

from aaf import logging
from aaf.event_handlers import use_event_handler
from aaf.llms import AnthropicRunner
from aaf.threads import Session
from aaf.tools import demo
from aaf.ui import SimpleStreamingUserInterface

DEFAULT_MODEL = "sonnet"

app = typer.Typer()
selected_model = DEFAULT_MODEL


@app.callback()
def main(
    model: Annotated[str, typer.Option("--model", "-m")] = DEFAULT_MODEL,
    quiet: Annotated[bool, typer.Option("--quiet", "-q")] = False,
    debug: Annotated[bool, typer.Option("--debug", "-d")] = False,
):
    global selected_model

    logging.set_level_from_flags(quiet=quiet, debug=debug)
    selected_model = model


async def run_question(question: str):
    thread = Session().create_thread(model=selected_model)
    thread.add_message("user", question)

    async with thread.run() as stream:
        async for chunk in stream.text_chunks():
            print(chunk.content, end="", flush=True)
        print()

    print(thread.cost_and_usage().pretty())


@app.command()
def question(question: str = "What's the capital of France?"):
    asyncio.run(run_question(question))


async def run_question_loop(question: str):
    thread = Session().create_thread(model=selected_model)
    thread.add_message("user", question)

    async with thread.run_loop() as stream:
        async for chunk in stream.text_chunks():
            print(chunk.content, end="", flush=True)
        print()

    print(thread.cost_and_usage().pretty())


@app.command()
def question_loop(question: str = "What's the capital of France?"):
    asyncio.run(run_question_loop(question))


async def run_pirate_multistep():
    system = (
        "Talk like a pirate. Answer briefly, without reflecting the question. Use tools only if explicitly instructed"
    )
    ui = SimpleStreamingUserInterface()
    thread = Session().create_thread(selected_model, system=system, tools=[demo.text_complexity])

    thread.add_message("user", "What is the meaning of life, in one short sentence?")
    async with thread.run(name="first_question") as stream:
        await use_event_handler(stream, ui)

    thread.add_message("user", "Print the previous answer in ALL CAPS")
    async with thread.run_loop() as stream:
        await use_event_handler(stream, ui)

    thread.add_message("user", "Calculate the text complexity of the previous answer. Use tools provided")
    async with thread.run_loop() as stream:
        await use_event_handler(stream, ui)

    print(thread.cost_and_usage().pretty())


@app.command()
def pirate_multistep():
    asyncio.run(run_pirate_multistep())


async def run_weather():
    tools: list[Callable] = [demo.get_weather_at, demo.get_location_coordinates]

    ui = SimpleStreamingUserInterface()
    thread = Session().create_thread(selected_model, tools=tools)

    thread.add_message("user", "What's the current weather in Tallinn, New York City, and Paris? ")
    async with thread.run_loop() as stream:
        await use_event_handler(stream, ui)

    thread.add_message("user", "Briefly (no more than 3 lines) aggregate results across locations.")
    async with thread.run_loop() as stream:
        await use_event_handler(stream, ui)

    ui.info(f"Final cost: USD {thread.cost_and_usage().cost:.4f}")


@app.command()
def weather():
    asyncio.run(run_weather())


async def run_rate_limits():
    model = "claude-3-5-sonnet-20240620"
    messages = [
        {
            "role": "user",
            "content": "hi",
        },
    ]

    runner = AnthropicRunner()
    response = await runner.client.messages.with_raw_response.create(
        model=model,
        messages=messages,
        max_tokens=1,
    )

    for k, v in response.headers.items():
        if k.startswith("anthropic-ratelimit-"):
            print(f"{k}: {v}")


@app.command()
def rate_limits():
    asyncio.run(run_rate_limits())


@app.command()
def all():
    global selected_model

    commands = [
        question,
        question_loop,
        weather,
        pirate_multistep,
    ]
    models = [
        "gpt-4o-mini",
        "sonnet",
    ]

    for command in commands:
        for model in models:
            print()
            print()
            rich.print(f"[bright_yellow]RUNNING: {command.__name__} with {model}[/bright_yellow]")

            selected_model = model
            command()


if __name__ == "__main__":
    app()



import os
from typing import Optional

from .openai import OpenAIRunner
from .types import ModelInfo


class OllamaRunner(OpenAIRunner):
    MODELS = [
        ModelInfo(
            name="llama3.1:8b",
            aliases=["llama3.1", "llama3", "llama"],
        ),
    ]

    def __init__(self, *, base_url: Optional[str] = None, **kwargs):
        if base_url is None:
            base_url = os.environ.get("OLLAMA_BASE_URL")
        if base_url is None:
            base_url = f"http://localhost:11434/v1"

        super().__init__(base_url=base_url, **kwargs)



import asyncio
from typing import Annotated

import typer

from aaf import logging
from aaf.threads import Session
from aaf.virtual_models.multiphase import MultiphaseModel

DEFAULT_MODEL = MultiphaseModel.external_model_name


async def multiphase(model_name: str, question: str) -> None:
    runner = MultiphaseModel(model_name)
    # runner = SimpleVirtualModel()
    thread = Session().create_thread(model=runner.id, runner=runner)
    thread.add_message("user", question)

    async with thread.run() as stream:
        async for chunk in stream.text_chunks():
            print(chunk.content, end="", flush=True)
        print()

    print(thread.cost_and_usage().pretty())


async def main(model: str, question: str):
    await multiphase(model_name=model, question=question)


def main_sync(
    question: str,
    model: Annotated[str, typer.Option("--model", "-m")] = DEFAULT_MODEL,
    quiet: Annotated[bool, typer.Option("--quiet", "-q")] = False,
    debug: Annotated[bool, typer.Option("--debug", "-d")] = False,
):
    logging.set_level_from_flags(quiet=quiet, debug=debug)

    asyncio.run(main(model=model, question=question))


if __name__ == "__main__":
    typer.run(main_sync)



import asyncio
import json
from contextlib import asynccontextmanager
from datetime import datetime
from typing import AsyncIterator, Callable, Optional

from .forwarding import ResponseQueue, VirtualModelAdapter
from .llms import ModelRunner, get_llm_provider_and_model
from .llms.base import APIConnectionError, ModelResponseStream, RateLimitError
from .llms.types import (
    ChatRequest,
    CompletionUsage,
    CostAndUsage,
    ResponseChunkToolCall,
    ResponseChunkToolCallFailed,
    ResponseChunkToolCallFinished,
    ResponseChunkToolCallStarted,
    ResponseControlChunkRateLimited,
    ResponseDebugChunk,
    ResponseStopReasonChunk,
    ResponseStream,
    ToolCall,
    ToolCallResult,
)
from .logging import log
from .utils import truncate_text


class StatsTracking:
    """Helper base class that provides cost&usage and elapsed time tracking."""

    def __init__(self):
        super().__init__()

        self._started_at = datetime.now()

    def cost_and_usage(self) -> CostAndUsage:
        raise NotImplementedError()

    def elapsed_secs(self, now: Optional[datetime] = None) -> float:
        if now is None:
            now = datetime.now()

        return (now - self._started_at).total_seconds()


class Thread(StatsTracking):
    """Runs and represents a conversation with a model.

    Thread has methods to run either one step of a conversation (i.e. it gets one chat completion from the model),
    or it can run a loop which handles tool calls and multiple model completions.

    It also tracks cost and token usage.

    For actually running the model, it invokes the ModelRunner, asking it to run specified model
    with the given messages & parameters.
    """

    def __init__(
        self,
        model: str,
        request: ChatRequest,
        name: Optional[str] = None,
        *,
        runner: Optional[ModelRunner] = None,
        system: Optional[str] = None,
        tools: Optional[list[Callable]] = None,
    ):
        super().__init__()

        if runner is None:
            runner_cls, self.model = get_llm_provider_and_model(model)
            self.model_runner = runner_cls()
        else:
            self.model = model
            self.model_runner = runner

        self.name = name or "thread"
        self.request = request
        self.tools = {func.__name__: func for func in (tools or [])}

        self.messages = request.messages.copy()
        if system:
            self.messages.insert(0, {"role": "system", "content": system})

        self.runs_count = 0
        self.per_run_costs = []

    @asynccontextmanager
    async def run(
        self, *, name: Optional[str] = None, ignore_exceptions: bool = False, **kwargs
    ) -> AsyncIterator[ResponseStream]:
        log.info("Thread.run()")

        response = ResponseQueue()
        process_task = asyncio.create_task(self._run_step_background(response, name=name, **kwargs))
        stream = ModelResponseStream(VirtualModelAdapter(process_task, response))

        log.debug("Thread.run(): yield stream")
        yield stream

        await stream.finish(ignore_exceptions)
        log.info("Thread.run(): DONE")

    async def _run_step_background(self, response: ResponseQueue, *, name: Optional[str] = None, **kwargs) -> None:
        log.info("Thread._run_step_background()")

        if name is None:
            name = f"run_{self.runs_count}"

        request = self.request.model_copy(update={"messages": self.messages, "tools": list(self.tools.values())})

        # The loop handles retries for rate limits
        max_retries = 5
        for step in range(max_retries + 1):
            try:
                stream = await self.model_runner.run(self.model, request, **kwargs)
                break

            except RateLimitError as error:
                log.warning("Rate limit exceeded", step=step, error=error)
                if step >= max_retries:
                    await response.put(error)
                    await response.mark_finished()
                    return

                log.info("Rate limit recovery - waiting", step=step, seconds=error.retry_in_secs)
                await response.put(
                    ResponseControlChunkRateLimited(delay_secs=error.retry_in_secs, metadata=error.metadata)
                )
                await asyncio.sleep(error.retry_in_secs)

            except APIConnectionError as error:
                log.warning("API connection error", step=step, error=error)
                await asyncio.sleep(0.5)

            except Exception as error:
                # This will cause the user code to freeze, since the `yield` will not be executed.
                #  There are ways to fix it, but they're complicated, so for now we just print out the alert for the user.
                log.error(
                    f"ERROR: opening stream failed. Code will freeze!", model=self.model, runner=self.model_runner
                )
                log.critical(f"ERROR: {error}")
                raise

        else:
            log.error("Thread._run_step_background(): reached maximum number of iterations")
            await response.mark_finished()
            return

        try:
            async for chunk in stream:
                await response.put(chunk)

            # Ensure we've processed the full stream, even if the user didn't
            await stream.finish()
            await response.mark_finished()

            for msg in self.model_runner.get_assistant_messages(stream):
                self.add_message(**msg)

        finally:
            self.per_run_costs.append(self.create_cost_and_usage(name, stream.usage))
            self.runs_count += 1

    @asynccontextmanager
    async def run_loop(
        self, *, max_iterations: int = 10, ignore_exceptions: bool = False
    ) -> AsyncIterator[ResponseStream]:
        log.info("Thread.run_loop()", max_iterations=max_iterations)

        response = ResponseQueue()
        process_task = asyncio.create_task(self._run_loop_background(response, max_iterations=max_iterations))
        stream = ModelResponseStream(VirtualModelAdapter(process_task, response))

        log.debug("Thread.run_loop(): yield stream")
        yield stream

        await stream.finish(ignore_exceptions)
        log.info("Thread.run_loop(): DONE")

    async def _run_loop_background(self, response: ResponseQueue, max_iterations: int) -> None:
        log.debug("Thread.run_loop_background()", max_iterations=max_iterations)
        for i in range(max_iterations):
            tool_use_requests: list[ToolCall] = []
            async with self.run(ignore_exceptions=True) as step_stream:
                log.debug("Thread._run_loop_background(): stream", stream_id=id(step_stream))
                async for chunk in step_stream:
                    log.debug("Thread.run_loop_background(): got chunk", chunk=chunk)
                    if isinstance(chunk, Exception):
                        log.error("Thread.run_loop_background(): got exception", exception=chunk)
                        await response.put(chunk)
                        await response.mark_finished()
                        return

                    elif isinstance(chunk, ResponseChunkToolCall):
                        tool_use_requests.append(chunk.tool_call)
                    elif isinstance(chunk, ResponseStopReasonChunk):
                        log.info("Thread.run_loop_background(): got stop reason", reason=chunk.reason)
                    else:
                        # log.debug("Thread.run_loop_background(): putting chunk", chunk=chunk)
                        await response.put(chunk)
                log.debug("Thread._run_loop_background(): stream done", stream_id=id(step_stream))

            log.debug("Thread.run_loop_background(): loop finished")
            if not tool_use_requests:
                break

            # execute tool use requests
            await response.put(ResponseDebugChunk(content="Executing tool use requests"))
            results = await self.execute_tool_calls(tool_use_requests, response)
            for message in self.model_runner.get_tool_result_messages(results):
                self.add_message(**message)

        await response.mark_finished()

    def create_cost_and_usage(self, name: str, usage: CompletionUsage) -> CostAndUsage:
        return CostAndUsage(
            name=name,
            completion_tokens=usage.completion_tokens,
            prompt_tokens=usage.prompt_tokens,
            cost=self.model_runner.cost_in_usd(self.model, usage),
        )

    def cost_and_usage(self) -> CostAndUsage:
        return CostAndUsage.aggregate(self.name, self.per_run_costs)

    def add_message(self, role: str, content: str, **kwargs) -> None:
        self.messages.append({"role": role, "content": content, **kwargs})

    async def execute_tool_calls(self, tool_calls: list[ToolCall], queue: ResponseQueue) -> list[ToolCallResult]:
        """Executes the tool calls generated by the model.

        Message containing the requested tool calls is added to the stream.
        Another message is added to the stream for each tool call result.
        """

        # self.event_handler.debug(f"Executing {len(tool_calls)} tool calls...")
        results = await asyncio.gather(*[self.execute_tool_call(tool_call, queue) for tool_call in tool_calls])
        return list(results)

    async def execute_tool_call(self, tool_call: ToolCall, queue: ResponseQueue) -> ToolCallResult:
        """Executes a single tool call, which can be async or sync."""

        log.info("Thread.execute_tool_call()", name=tool_call.name, arguments=tool_call.arguments)
        try:
            await queue.put(ResponseChunkToolCallStarted(tool_call=tool_call))

            tool = self.tools[tool_call.name]
            params = tool_call.arguments_dict.copy()

            if asyncio.iscoroutinefunction(tool):
                result = await tool(**params)
            else:
                result = tool(**params)

        except Exception as error:
            log.warning("Thread.execute_tool_call(): error", name=tool_call.name, arguments=tool_call.arguments)
            await queue.put(ResponseChunkToolCallFailed(tool_call=tool_call, error=repr(error)))
            return ToolCallResult(tool_call=tool_call, is_error=True, content="Error")
        else:
            log.info(
                "Thread.execute_tool_call(): result",
                name=tool_call.name,
                arguments=tool_call.arguments,
                result=truncate_text(result, 500) if isinstance(result, str) else result,
            )
            await queue.put(ResponseChunkToolCallFinished(tool_call=tool_call, result=result))
            return ToolCallResult(tool_call=tool_call, content=json.dumps(result))


class Session(StatsTracking):
    """Session keeps configuration and tracks costs.

    It's usually the main entrypoint for creating threads, and thus starting conversations.
    """

    def __init__(self, name: str = "root"):
        super().__init__()

        self.name = name
        self.config: dict = {}

        self.threads = []

    def create_thread(
        self,
        model: str,
        request: Optional[ChatRequest] = None,
        name: Optional[str] = None,
        *,
        runner: Optional[ModelRunner] = None,
        system: Optional[str] = None,
        tools: Optional[list[Callable]] = None,
    ) -> Thread:
        if request is None:
            request = ChatRequest()

        thread = Thread(model=model, request=request, name=name, runner=runner, system=system, tools=tools)
        self.threads.append(thread)
        return thread

    def cost_and_usage(self) -> CostAndUsage:
        return CostAndUsage.aggregate(self.name, [thread.cost_and_usage() for thread in self.threads])



import asyncio
from typing import Callable

from ..forwarding import ResponseQueue
from ..llms.types import ChatRequest
from ..threads import Session
from ..tools import demo
from .base import VirtualModelBase


class MinimalVirtualModel(VirtualModelBase):
    id: str = "rivo/minimal-proxy"

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        await asyncio.sleep(0.1)
        await queue.add("hello ")
        await asyncio.sleep(0.1)
        await queue.add("there!")
        await asyncio.sleep(1)
        await queue.add(" I am ")
        await asyncio.sleep(0.1)
        await queue.add(self.id)
        await asyncio.sleep(0.1)
        await queue.add("!")
        await queue.add("\n")
        await queue.add(
            """<details open>
            <summary>Hello</summary>
            World!
            </details>"""
        )


class SimpleVirtualModel(VirtualModelBase):
    id: str = "rivo/simple-proxy"

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        thread = Session().create_thread("gpt-4o", system="Pretend to be a pirate")
        for message in chat.messages:
            thread.add_message(**message)

        async with thread.run() as stream:
            await stream.text_chunks().redirect(queue)

        await self.output_addendum(queue, thread)


class DemoToolUsageModel(VirtualModelBase):
    id: str = "rivo/demo-tools"

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        tools: list[Callable] = [demo.get_weather_at, demo.get_location_coordinates]
        thread = Session().create_thread("gpt-4o", tools=tools)

        for message in chat.messages:
            thread.add_message(**message)

        async with thread.run_loop() as stream:
            await stream.text_chunks().redirect(queue)

        await self.output_addendum(queue, thread)



from typing import AsyncIterator, Callable, Optional

import openai
from openai import NOT_GIVEN, AsyncOpenAI, AsyncStream
from openai.types.chat import ChatCompletionChunk
from openai.types.completion_usage import CompletionUsage as OpenAICompletionUsage

from ..logging import log
from ..tools_support.schema import jsonschema_for_function
from .base import APIConnectionError, ModelResponseStream, ModelRunner, ResponseAdapterBase
from .types import (
    ChatRequest,
    CompletionUsage,
    ModelCost,
    ModelInfo,
    ResponseChunk,
    ResponseChunkToolCall,
    ResponseStopReasonChunk,
    ResponseTextChunk,
    ResponseUsageChunk,
    StopReason,
    ToolCall,
    ToolCallResult,
)


class OpenAIResponseAdapter(ResponseAdapterBase):
    def __init__(self, stream: AsyncStream[ChatCompletionChunk]):
        self.stream = stream

        self.raw_chunks = []

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        log.debug("OpenAIResponseAdapter.__aiter__()", id=id(self))
        async for chunk in self.stream:
            log.debug("OpenAIResponseAdapter.__aiter__(): got raw chunk", chunk=chunk)
            self.raw_chunks.append(chunk)

            if chunk.choices and chunk.choices[0].delta and chunk.choices[0].delta.content:
                content = chunk.choices[0].delta.content
                yield ResponseTextChunk(content=content)

            if chunk.usage is not None:
                yield ResponseUsageChunk(delta=self.parse_usage(chunk.usage))

            if chunk.choices and chunk.choices[0].finish_reason:
                # Emit any and all tool calls
                for tool_call in self.get_tool_calls():
                    yield ResponseChunkToolCall(tool_call=tool_call)

                reason = self.parse_stop_reason(chunk.choices[0].finish_reason)
                yield ResponseStopReasonChunk(reason=reason)

        log.debug("OpenAIResponseAdapter.__aiter__(): done", id=id(self))

    def parse_usage(self, usage: OpenAICompletionUsage) -> CompletionUsage:
        return CompletionUsage(prompt_tokens=usage.prompt_tokens, completion_tokens=usage.completion_tokens)

    def parse_stop_reason(self, openai_reason: Optional[str]) -> Optional[StopReason]:
        if openai_reason == "stop":
            return StopReason.END_TURN
        if openai_reason == "tool_calls":
            return StopReason.TOOL_USE
        if openai_reason == "length":
            return StopReason.LENGTH

        return None

    def get_tool_calls(self) -> list[ToolCall]:
        calls: list[ToolCall] = []

        # Iterate through each chunk
        chunk: ChatCompletionChunk
        for chunk in self.raw_chunks:
            if not chunk.choices:
                continue

            assert len(chunk.choices) == 1
            choice = chunk.choices[0]  # Assumes one choice for simplicity

            # Check if the choice has tool calls
            if not choice.delta.tool_calls:
                continue
            for tool_call in choice.delta.tool_calls:
                idx = tool_call.index
                if idx >= len(calls):
                    calls.append(ToolCall(id="", name="", arguments=""))

                if tool_call.function:
                    # Should id & name be appended or replaced?
                    if tool_call.id:
                        calls[idx].id = tool_call.id
                    if tool_call.function.name:
                        calls[idx].name = tool_call.function.name
                    if tool_call.function.arguments:
                        calls[idx].arguments += tool_call.function.arguments

        return calls


class OpenAIRunner(ModelRunner):
    MODELS = [
        ModelInfo(
            name="gpt-4o-2024-08-06",
            aliases=["gpt-4o", "4o", "gpt4", "gpt"],
            cost=ModelCost(prompt_per_1m=2.50, completion_per_1m=10),
        ),
        ModelInfo(
            name="gpt-4o-mini-2024-07-18",
            aliases=["gpt-4o-mini", "gpt-mini"],
            cost=ModelCost(prompt_per_1m=0.15, completion_per_1m=0.60),
        ),
        ModelInfo(
            name="chatgpt-4o-latest",
            aliases=["chatgpt", "chat"],
            cost=ModelCost(prompt_per_1m=5, completion_per_1m=15),
        ),
    ]

    def __init__(self, **kwargs):
        super().__init__()
        self.client = AsyncOpenAI(**kwargs)

    async def run(self, model: str, request: ChatRequest, **kwargs) -> ModelResponseStream:
        self.log_run_request(model, request, kwargs)

        # Ensure the model is valid
        self.get_model_info(model)

        try:
            stream = await self.client.chat.completions.create(
                model=model,
                messages=request.messages,
                stream=True,
                stream_options={"include_usage": True},
                tools=self.get_tools_schema(request.tools) if request.tools else NOT_GIVEN,
                **kwargs,
            )
        except openai.APIConnectionError as error:
            raise APIConnectionError(message=error.message) from error

        return ModelResponseStream(OpenAIResponseAdapter(stream))

    def get_tools_schema(self, tools: list[Callable]) -> list[dict]:
        return [{"type": "function", "function": jsonschema_for_function(func)} for func in tools]

    def get_tool_result_messages(self, results: list[ToolCallResult]) -> list[dict]:
        """OpenAI uses one message per tool call result."""

        return [
            {
                "role": "tool",
                "content": result.content if not result.is_error else "Error",
                "tool_call_id": result.tool_call.id,
                "name": result.tool_call.name,
            }
            for result in results
        ]

    def get_assistant_messages(self, stream: ModelResponseStream) -> list[dict]:
        if stream.stop_reason == StopReason.TOOL_USE:
            # Add assistant message with tool calls
            return [
                {
                    "role": "assistant",
                    "content": None,
                    "tool_calls": [
                        {
                            "id": tool_call.id,
                            "type": "function",
                            "function": {"name": tool_call.name, "arguments": tool_call.arguments},
                        }
                        for tool_call in stream.get_tool_calls()
                    ],
                }
            ]
        else:
            # Add assistant message with completion
            return [{"role": "assistant", "content": stream.text}]



from .. import prompts
from ..forwarding import ResponseQueue
from ..llms.types import ChatRequest
from ..logging import log
from ..threads import Session
from ..utils import extract_xml_fragment
from .base import VirtualModelBase


class TwoPhaseModel(VirtualModelBase):
    id: str = "rivo/two-phase"

    external_model_name = "sonnet-3.5"

    async def process(self, chat: ChatRequest, queue: ResponseQueue) -> None:
        session = Session()
        user_messages = [message for message in chat.messages if message["role"] == "user"]
        if len(user_messages) > 1:
            return await self.process_continuation(session, chat, queue, self.external_model_name)

        question = user_messages[0]["content"]

        prompt = await self.generate_prompt(session, queue, question)
        await queue.add_debug(f"## _TwoPhase: prompt_\n\n{prompt}\n\n---\n\n")

        await self.generate_answer(session, queue, question, prompt)

        await self.output_addendum(queue, session)

    async def generate_prompt(self, session: Session, queue: ResponseQueue, question: str) -> str:
        thread = session.create_thread(self.external_model_name, name="prompt")
        thread.add_message("system", prompts.PROMPT_ENGINEER)
        thread.add_message("user", question)

        async with self.run_progressbar(queue, message="Writing prompt"):
            async with thread.run() as stream:
                await stream.finish()

        full_response = thread.messages[-1]["content"]
        log.debug("TwoPhaseModel.generate_prompt(): Full response", response=full_response)

        result = extract_xml_fragment(full_response, "system_prompt")
        log.info("TwoPhaseModel.generate_prompt() done", result_len=len(result), cost=thread.cost_and_usage().as_log())
        return result

    async def generate_answer(self, session: Session, queue: ResponseQueue, question: str, prompt: str) -> None:
        thread = session.create_thread(self.external_model_name, name="answer")
        thread.add_message("system", prompt)
        thread.add_message("user", question)

        async with thread.run() as stream:
            await stream.text_chunks().redirect(queue)

        result = thread.messages[-1]["content"]
        log.info("TwoPhaseModel.generate_answer() done", result_len=len(result), cost=thread.cost_and_usage().as_log())
        return result


class TwoPhaseGPTModel(TwoPhaseModel):
    id: str = "rivo/two-phase-gpt"

    external_model_name = "gpt-4o"


class TwoPhaseChatGPTModel(TwoPhaseModel):
    id: str = "rivo/two-phase-chatgpt"

    external_model_name = "chatgpt-4o-latest"



import asyncio
from typing import Annotated

import typer

from aaf import logging, prompts
from aaf.event_handlers import use_event_handler
from aaf.threads import Session
from aaf.ui import SimpleStreamingUserInterface

DEFAULT_MODEL = "sonnet"


async def prompt_engineer(question: str, model_name: str):
    ui = SimpleStreamingUserInterface()
    thread = Session().create_thread(model_name)

    thread.add_message("system", prompts.PROMPT_ENGINEER)
    thread.add_message("user", question)
    async with thread.run_loop() as stream:
        await use_event_handler(stream, ui)

    ui.info(f"Final cost: USD {thread.cost_and_usage().cost:.4f}")


def cli(
    question: str,
    model: Annotated[str, typer.Option("--model", "-m")] = DEFAULT_MODEL,
    quiet: Annotated[bool, typer.Option("--quiet", "-q")] = False,
    debug: Annotated[bool, typer.Option("--debug", "-d")] = False,
):
    logging.set_level_from_flags(quiet=quiet, debug=debug)

    asyncio.run(prompt_engineer(question, model_name=model))


if __name__ == "__main__":
    typer.run(cli)



import asyncio
from asyncio import Queue
from typing import AsyncIterator

from .llms.base import ResponseAdapterBase
from .llms.types import (
    ResponseChunk,
    ResponseDebugChunk,
    ResponseStopReasonChunk,
    ResponseTextChunk,
    ResponseVerboseChunk,
    StopReason,
)


class ResponseQueue(Queue):
    """Queue for sending response chunks.

    This is similar to ResponseStream, but focuses on the writing part (whereas ResponseStream focuses on reading).
    It's a subclass of asyncio.Queue, with some convenience methods for adding different types of chunks.
    """

    async def add(self, content: str) -> None:
        await self.put(ResponseTextChunk(content=content))

    async def add_debug(self, content: str) -> None:
        await self.put(ResponseDebugChunk(content=content))

    async def add_verbose(self, content: str) -> None:
        await self.put(ResponseVerboseChunk(content=content))

    async def mark_finished(self) -> None:
        await self.put(None)


class VirtualModelAdapter(ResponseAdapterBase):
    """Response stream adapter for virtual models.

    It reads chunks from the provided queue and forwards them to the stream.
    The queue must end with a None value to indicate the end of the stream.
    """

    def __init__(self, process_task: asyncio.Task[None], queue: ResponseQueue):
        super().__init__()

        self.process_task = process_task
        self.queue = queue

        self._stop_reason = None

    async def __aiter__(self) -> AsyncIterator[ResponseChunk]:
        """Forward chunks from the queue to the stream.

        Takes care to avoid yielding the stop reason twice.
        """

        try:
            while True:
                chunk = await self.queue.get()
                if chunk is None:  # End of stream marker
                    if self._stop_reason is None:
                        yield ResponseStopReasonChunk(reason=StopReason.END_TURN)

                    break

                if isinstance(chunk, ResponseStopReasonChunk):
                    self._stop_reason = chunk.reason

                yield chunk

        finally:
            # Ensure the process task is completed
            await self.process_task



import asyncio

from duckduckgo_search import AsyncDDGS


async def search(query: str) -> list[dict]:
    """Search the web for the given query."""

    results = await AsyncDDGS().atext(query, max_results=10)
    return results


async def search_multi(queries: list[str]) -> list[list[dict]]:
    """Execute multiple web searches at once.

    Behaves same as "search" tool, but runs multiple queries in parallel.
    Returned results are in the same order as the input queries.
    """

    return await asyncio.gather(*[search(query) for query in queries])
